import os
import pytest

from aiofs.redis import RedisFileSystem, RedisConfig

FILENAME = "filename"


@pytest.fixture
async def client(request, redis_config):
    pattern = request.params if hasattr(request, "param") else "*"
    fs = RedisFileSystem(redis_config, pattern)
    yield fs
    await fs.rm(FILENAME)
    await fs.close()


# Parametrize the class
class A_RedisFileLike:

    async def should_persist_content(self, client):
        f = client.open(FILENAME, "w")
        async with f:
            await f.write(b"content")

        f = client.open(FILENAME)
        async with f:
            assert await f.read() == b"content"

    async def should_update_content(self, client):
        f = client.open(FILENAME, "w")
        async with f:
            await f.write(b"content")

        f = client.open(FILENAME, "r+")
        async with f:
            content = await f.read()
            await f.write(content + b"-with-append")

        f = client.open(FILENAME)
        async with f:
            assert await f.read() == b"content-with-append"

    async def should_raise_file_not_found(self, client):
        f = client.open(FILENAME)
        with pytest.raises(FileNotFoundError):
            async with f:
                await f.read()

    async def should_run_within_context(self, client):
        f = client.open(FILENAME)
        with pytest.raises(RuntimeError):
            await f.read()

        with pytest.raises(RuntimeError):
            await f.write(b"")

    async def should_list_files(self, client):
        async with client.open("one/file", "w") as f:
            await f.write(b"content")

        async with client.open("other_one", "w") as f:
            await f.write(b"content")

        async with client.open("the third", "w") as f:
            await f.write(b"content")

        assert set(await client.ls()) == set(["one/file", "other_one", "the third"])
        assert set(await client.ls("*ne*")) == set(["one/file", "other_one"])



from dataclasses import dataclass
import pytest
from testcontainers.core.container import DockerContainer
from testcontainers.core.wait_strategies import LogMessageWaitStrategy

from aiofs.redis import RedisConfig
from aiofs.azure import AzureContainerConfig


@pytest.fixture(scope="session")
def redis_container():
    """Start Redis container for the test session."""
    container = DockerContainer("redis:5.0.3-alpine")
    container.with_exposed_ports(6379)
    with container:
        strategy = LogMessageWaitStrategy("Ready to accept connections")
        strategy.with_startup_timeout(30)
        container.waiting_for(strategy)
        yield container


@pytest.fixture(scope="session")
def azurite_container():
    """Start Azurite container for the test session."""
    container = DockerContainer("mcr.microsoft.com/azure-storage/azurite:latest")
    container.with_exposed_ports(10000)
    container.with_command(
        "azurite --blobHost 0.0.0.0 --blobPort 10000 --skipApiVersionCheck"
    )
    with container:
        strategy = LogMessageWaitStrategy(
            "Azurite Blob service is successfully listening"
        )
        strategy.with_startup_timeout(30)
        container.waiting_for(strategy)
        yield container


@pytest.fixture(scope="session")
def redis_connection_params(redis_container):
    """Get Redis connection parameters."""
    return {
        "host": redis_container.get_container_host_ip(),
        "port": redis_container.get_exposed_port(6379),
    }


class AzuriteConfig(AzureContainerConfig):

    def __init__(self, container_name, host, port):
        self._host = host
        self._port = port
        super().__init__("", "", container_name)

    @property
    def connection_string(self) -> str:
        return (
            "AccountName=devstoreaccount1;"
            "AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/"
            "K1SZFPTOtr/KBHBeksoGMGw==;"
            "DefaultEndpointsProtocol=http;"
            f"BlobEndpoint=http://{self._host}:{self._port}/devstoreaccount1;"
        )


@pytest.fixture(scope="session")
def azure_config(azurite_container):
    host = azurite_container.get_container_host_ip()
    port = azurite_container.get_exposed_port(10000)
    return AzuriteConfig("aiofs-cnt", host, port)


@pytest.fixture(scope="session")
def redis_config(redis_container):
    host = redis_container.get_container_host_ip()
    port = redis_container.get_exposed_port(6379)

    return RedisConfig(host=host, port=port)



import os
import pytest

from aiofs.azure import AzFileSystem, AzureContainerConfig


HOSTNAME = os.getenv("AZURE_HOSTNAME", "127.0.0.1")
FILENAME = "filename"


@pytest.fixture
async def fs(azure_config):
    fs = AzFileSystem(azure_config, "prefix/*/posfix")
    yield fs
    await fs.rm("*")
    await fs.close()


# Parametrize the class
class An_AzFileLike:
    async def should_have_template_customizable(self, fs):

        fs.template = "{}"
        assert fs.template == "{}"

    async def should_persist_content(self, fs):
        f = fs.open(FILENAME, "w")
        async with f:
            await f.write(b"content")

        f = fs.open(FILENAME)
        async with f:
            assert await f.read() == b"content"

    async def should_update_content(self, fs):
        f = fs.open(FILENAME, "w")
        async with f:
            await f.write(b"content")

        f = fs.open(FILENAME, "r+")
        async with f:
            content = await f.read()
            await f.write(content + b"-with-append")

        f = fs.open(FILENAME)
        async with f:
            assert await f.read() == b"content-with-append"

    async def should_raise_file_not_found(self, fs):
        f = fs.open(FILENAME)
        with pytest.raises(FileNotFoundError):
            async with f:
                await f.read()

    async def should_run_within_context(self, fs):
        f = fs.open(FILENAME)
        with pytest.raises(RuntimeError):
            await f.read()

        with pytest.raises(RuntimeError):
            await f.write(b"")

    async def should_list_files(self, fs):
        async with fs.open("one/file", "w") as f:
            await f.write(b"content")

        async with fs.open("other_one", "w") as f:
            await f.write(b"content")

        async with fs.open("the third", "w") as f:
            await f.write(b"content")

        assert set(await fs.ls()) == set(["one/file", "other_one", "the third"])
        assert set(await fs.ls("*ne*")) == set(["one/file", "other_one"])

    async def should_list_files_without_suffix(self, fs):
        fs.template = "{}"
        async with fs.open("one/file", "w") as f:
            await f.write(b"content")

        async with fs.open("other_one", "w") as f:
            await f.write(b"content")

        async with fs.open("the third", "w") as f:
            await f.write(b"content")

        assert set(await fs.ls()) == set(["one/file", "other_one", "the third"])
        assert set(await fs.ls("*ne*")) == set(["one/file", "other_one"])



"""Implementation of FileLike in a real FileSystem"""
import os

import aiofiles

from . import FileLike, FileLikeSystem
from fnmatch import fnmatch


class OsFile(FileLike):
    """Implementation of FileLike in a real FileSystem"""

    def __init__(self, filename: str, mode: str):
        self._fn = filename
        self._mode = mode if "b" in mode else "b" + mode
        self._context = None
        self._f = None
        self._new = True
        self._content = b""

    async def __aenter__(self):
        if os.path.exists(self._fn):
            self._context = aiofiles.open(self._fn, self._mode)
            self._f = await self._context.__aenter__()
            self._new = False
        else:
            self._context = aiofiles.open(self._fn, "bw")
            self._f = await self._context.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._context and self._f:
            if self._content:
                await self._f.seek(0)
                await self._f.write(self._content)
                self._content = b""

            await self._context.__aexit__(exc_type, exc, tb)
            self._context = None
            self._f = None

    async def write(self, b: bytes):
        if not self._f:
            raise RuntimeError(f"Method used out of a context")

        self._content += b

    async def read(self, size: int = -1) -> bytes:
        if not self._f:
            raise RuntimeError(f"Method used out of a context")

        if self._new:
            raise FileNotFoundError(self._fn)

        return await self._f.read(size)


class OsFileSystem(FileLikeSystem):
    """Implementation"""

    def __init__(self, template: str = "/{}"):
        self._template = template

    def _get_full_fn(self, filename: str) -> str:
        return self._template.format(filename)

    def open(self, filename: str, mode: str = "r") -> OsFile:
        full_fn = self._get_full_fn(filename)
        os.makedirs(os.path.dirname(full_fn), exist_ok=True)
        return OsFile(full_fn, mode)

    @property
    def template(self) -> str:
        """Template to generate the full filename"""
        return self._template

    @template.setter
    def template(self, value: str):
        """Change template for locating the files"""
        self._template = value

    async def rm(self, pattern: str):
        fns = await self.ls(pattern)
        for fn in fns:
            os.remove(self._get_full_fn(fn))

    async def ls(self, pattern: str = "*"):
        prefix, _, sufix = self._template.partition("{}")
        pre = 0 if prefix is None else len(prefix)
        suf = -len(sufix) if sufix else None

        values = []
        for root, _, files in os.walk(prefix):
            for file in files:
                full_name = os.path.join(root, file)
                filename = full_name[pre:suf]
                if fnmatch(filename, pattern):
                    values.append(filename)

        return values



"""Implementation of FileLike with azure blobs"""

import fnmatch
from dataclasses import dataclass
from typing import Self

from azure.core.exceptions import ResourceExistsError, ResourceNotFoundError
from azure.storage.blob import BlobServiceClient as SyncBlobServiceClient
from azure.storage.blob.aio import BlobClient, BlobServiceClient

from . import FileLike, FileLikeSystem


@dataclass
class AzureContainerConfig:
    account_name: str
    account_key: str
    container_name: str
    protocol: str = "https"
    endpoint_suffix: str = "core.windows.net"

    @property
    def connection_string(self) -> str:
        """Dynamically generates the Azure Storage Connection String"""
        return (
            f"DefaultEndpointsProtocol={self.protocol};"
            f"AccountName={self.account_name};"
            f"AccountKey={self.account_key};"
            f"EndpointSuffix={self.endpoint_suffix}"
        )


class AzFileLike(FileLike):
    """Implementation of FileLike in azure blobs"""

    def __init__(self, blob: BlobClient, mode: str = "r"):
        self._blob = blob
        self._lease = None
        self._content = b""
        self._mode = mode
        self._lease = None
        self._context = None

    async def __aenter__(self) -> Self:
        self._context = self._blob = await self._blob.__aenter__()
        if "w" in self._mode or "r+" in self._mode:
            try:
                self._lease = await self._blob.acquire_lease(15)
            except ResourceNotFoundError:
                ...
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._blob.__aexit__(exc_type, exc, tb)

        try:
            if self._content:
                await self._blob.upload_blob(
                    self._content, overwrite=True, lease=self._lease
                )
        except ResourceExistsError:
            raise BlockingIOError(f"")
        finally:
            if self._lease:
                await self._lease.release()

    async def read(self, size: int = -1) -> bytes:
        if not self._context:
            raise RuntimeError('Call to "read" out of context')
        if size >= 0:
            raise NotImplementedError("Not designed to donwload partially")

        try:
            stream = await self._blob.download_blob()
        except ResourceNotFoundError:
            raise FileNotFoundError(f"Not found {self._blob.blob_name}")
        return await stream.readall()

    async def write(self, b: bytes):
        if not self._context:
            raise RuntimeError('Call to "write" out of context')
        self._content += b


class AzFileSystem(FileLikeSystem):
    """Create and manage AzFileLike"""

    def __init__(
        self,
        config: AzureContainerConfig,
        file_pattern: str = "*",
    ):
        print(config.connection_string)
        self._assert_container_exists(config)
        self._service = BlobServiceClient.from_connection_string(
            config.connection_string
        )
        self._container = self._service.get_container_client(config.container_name)

        self._pattern = file_pattern

    def open(self, filename: str, mode="r") -> AzFileLike:
        blob_name = self._pattern.replace("*", filename)
        blob = self._container.get_blob_client(blob_name)

        return AzFileLike(blob, mode)

    async def rm(self, pattern: str):
        filenames = await self.ls(pattern)
        for fn in filenames:
            blob_name = self._pattern.replace("*", fn)
            blob = self._container.get_blob_client(blob_name)
            try:
                await blob.delete_blob()
            except ResourceNotFoundError:
                pass

    async def close(self):
        if self._container:
            await self._container.close()
        await self._service.close()

    async def ls(self, pattern: str = "*"):
        """List all filenames of the system"""
        prefix, _, sufix = self._pattern.partition("*")
        names = []
        prefix = prefix if prefix else None
        pre = 0 if prefix is None else len(prefix)
        suf = -len(sufix) if sufix else None
        async for blob in self._container.list_blobs(prefix):
            if blob.name.endswith(sufix):
                names.append(blob.name[pre:suf])
        return fnmatch.filter(names, pattern)

    def _assert_container_exists(self, cfg: AzureContainerConfig):
        with SyncBlobServiceClient.from_connection_string(
            cfg.connection_string
        ) as service:
            if not service.get_container_client(cfg.container_name).exists():
                service.create_container(cfg.container_name)



from typing import Self, Protocol


class FileLike(Protocol):
    """Interface to be implemented by storage systems like a asynchronous file"""

    async def __aenter__(self) -> Self:
        """Context manager"""
        ...

    async def __aexit__(self, exc_type, exc, tb):
        """Context manager"""
        ...

    async def read(self, size: int = -1) -> bytes:
        """Read content of file"""
        ...

    async def write(self, b: bytes):
        """Write bytes to the file"""
        ...

    # async def close(self):
    #     """Close file"""

    # async def open(self):
    #     """Open the file"""

    # async def seekable(self) -> bool:
    #     return True

    # close
    # flush
    # isatty
    # read
    # readall
    # read1
    # readinto
    # seek
    # seekable
    # tell
    # truncate
    # writable
    # write
    # writelines
    # readline
    # readlines


class FileLikeSystem(Protocol):
    def open(self, filename: str, mode: str) -> FileLike:
        """Create a FileLike ready to be used"""
        ...

    async def rm(self, pattern: str):
        """Remove file-like object"""
        ...

    async def ls(self, pattern: str = "*") -> list[str]:
        """List all filenames of system with simple pattern (* and ?)"""
        ...

    async def mv(self, origin_path: str, destination_path: str):
        """Move a file to other location"""

    async def cp(self, origin_path: str, destination_path: str):
        """Copy a file to other path"""



import pytest

from aiofs.os import OsFileSystem


@pytest.fixture
def client(tmp_path):
    local_dir = tmp_path / "dir"
    template = f"{local_dir}/{{}}/file"

    return OsFileSystem(template)


class A_OsFile:
    async def should_have_template_customizable(self, client):
        client.template = "{}"
        assert client.template == "{}"

    async def should_persist_content(self, client):
        f = client.open("dir/file", "w")
        async with f:
            await f.write(b"content")

        f = client.open("dir/file")
        async with f:
            assert await f.read() == b"content"

    async def should_update_content(self, client):
        f = client.open("dir/file", "w")
        async with f:
            await f.write(b"content")

        f = client.open("dir/file", "r+")
        async with f:
            content = await f.read()
            await f.write(content + b"-with-append")

        f = client.open("dir/file")
        async with f:
            assert await f.read() == b"content-with-append"

    async def should_raise_file_not_found(self, client):
        f = client.open("dir/file")
        with pytest.raises(FileNotFoundError):
            async with f:
                await f.read()

    async def should_run_within_context(self, client):
        f = client.open("dir/file")
        with pytest.raises(RuntimeError):
            await f.read()

        with pytest.raises(RuntimeError):
            await f.write(b"")

    async def should_remove_file(self, client):
        async with client.open("dir/file", "w") as f:
            await f.write(b"content")

        await client.rm("dir/file")

        with pytest.raises(FileNotFoundError):
            async with client.open("dir/file") as f:
                await f.read()

    async def should_list_files(self, client):
        async with client.open("one/file", "w") as f:
            await f.write(b"content")

        async with client.open("other_one", "w") as f:
            await f.write(b"content")

        async with client.open("the third", "w") as f:
            await f.write(b"content")

        assert set(await client.ls()) == set(["one/file", "other_one", "the third"])
        assert set(await client.ls("*ne*")) == set(["one/file", "other_one"])



import logging
from dataclasses import dataclass
from typing import Any

from redis import WatchError
from redis.asyncio import ConnectionPool, Redis

from . import FileLike, FileLikeSystem

logger = logging.getLogger(__name__)


@dataclass
class RedisConfig:
    host: str | None = None
    port: int | None = None
    protocol: str = "redis"
    username: str | None = None
    password: str | None = None
    vhost: str | None = None
    query: str = ""

    def uri(self):
        value = f"{self.protocol}://"
        if self.username:
            value += self.username
        if self.password:
            value += f":{self.password}"

        if value[-2:] != "//":
            value += "@"

        if self.host:
            value += self.host

        if self.port:
            value += f":{self.port}"

        if self.vhost:
            value += f"/{self.vhost}"

        if self.query:
            value += f"?{self.query}"
        return value


class RedisFileLike(FileLike):
    """Implementation of FileLike using Redis Database"""

    def __init__(
        self, pool: ConnectionPool, path: str, mode="r", expiry: None | int = None
    ):
        self._key = path.replace("/", ":")
        self._pool = pool
        self._mode = mode

        self._context = None
        self._conn = None
        self._content = b""
        self._expiry = expiry

        self._pipe_ctx = None
        self._pipe = None

    async def __aenter__(self):
        await self._open()
        return self

    async def _open(self):
        self._context = Redis(connection_pool=self._pool)
        self._conn = await self._context.__aenter__()
        if "w" in self._mode or "r+" in self._mode:
            self._pipe_ctx = self._conn.pipeline(transaction=True)
            self._pipe = await self._pipe_ctx.__aenter__()
            await self._pipe.watch(self._key)

    async def __aexit__(self, exc_type, exc, tb):
        try:
            if self._pipe and self._pipe_ctx:
                self._pipe.multi()
                if self._content:
                    await self._pipe.set(self._key, self._content, self._expiry)
                await self._pipe.execute()
                await self._pipe_ctx.__aexit__(exc_type, exc, tb)
                self._pipe = None
                self._pipe_ctx = None
                self._content = b""
        except WatchError:
            fn = self._key.replace(":", "/")
            raise BlockingIOError(f"File '{fn}' is open by other process")
        finally:
            if self._context:
                await self._context.__aexit__(exc_type, exc, tb)
            self._context = None

    async def read(self, size: int = -1) -> bytes:
        """Read bytes from the register"""
        if self._conn is None:
            raise RuntimeError('Method "read" called out of context')

        if size < 0:
            content = await self._conn.get(self._key)
            print(content)
            if content is None:
                raise FileNotFoundError(f"Not found {self._key} at Redis")
            return content

        return b""

    async def write(self, b: bytes):
        """Write content to redis register"""
        if self._conn is None:
            raise RuntimeError('Method "write" called out of context')

        self._content += b


class RedisFileSystem(FileLikeSystem):
    """Implement a file system to get access to FileLike objects"""

    def __init__(
        self, cfg: RedisConfig, file_pattern: str = "*", expiry_s: int | None = None
    ):
        self._pool = ConnectionPool.from_url(
            cfg.uri(),
            health_check_interval=10,
            socket_timeout=10,
            socket_keepalive=True,
            socket_connect_timeout=10,
            retry_on_timeout=True,
        )
        self._pattern = file_pattern
        self._expiry = expiry_s

    def open(self, filename: str, mode: str = "r") -> RedisFileLike:
        """Return FileLike object"""
        return RedisFileLike(
            self._pool, self._pattern.replace("*", filename), mode, self._expiry
        )

    async def rm(self, pattern: str):
        fns = await self.ls(pattern)
        async with Redis(connection_pool=self._pool) as conn:
            for fn in fns:
                await conn.delete(self._pattern.replace("*", fn).replace("/", ":"))

    async def ls(self, pattern: str = "*"):
        """List all filenames following a file pattern"""
        prefix, _, sufix = self._pattern.replace("/", ":").partition("*")
        pattern = self._pattern.replace("*", pattern).replace("/", ":")
        keys = []
        async with Redis(connection_pool=self._pool) as conn:
            cursor = 0
            while True:
                # Use the SCAN command to find keys matching the pattern
                cursor, batch = await conn.scan(cursor, match=pattern, count=100)

                keys.extend(
                    [name[len(prefix) : -len(sufix) or None].decode() for name in batch]
                )
                if cursor == 0:  # Cursor 0 means the scan is complete
                    break
        return [key.replace(":", "/") for key in keys]

    async def cp(self, origin_path: str, destination_path: str):
        """Copy a filename in origin_path to destination_path"""
        return await super().cp(origin_path, destination_path)

    async def mv(self, origin_path: str, destination_path: str):
        """Move or rename the"""
        await self.cp(origin_path, destination_path)
        await self.rm(origin_path)

    async def close(self):
        await self._pool.disconnect()



from io import UnsupportedOperation
from typing import Self

from . import FileLike, FileLikeSystem
from fnmatch import fnmatch


class InMemFile(FileLike):
    def __init__(self, name: str, mode: str, content: bytes | None = None):
        self._name = name
        self._mode = mode
        self._content = content

        self._context = None

    async def __aenter__(self) -> Self:
        self._context = self._name
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._context = None

    async def read(self, size: int = -1) -> bytes:
        if "r" not in self._mode:
            raise UnsupportedOperation(f"File {self._name} not readable")

        if self._content is None:
            raise FileNotFoundError(f"No such file or directory: '{self._name}'")

        return self._content

    async def write(self, b: bytes):
        if "+" not in self._mode and "w" not in self._mode:
            raise UnsupportedOperation(f"File {self._name} not writable")

        self._content = b


class InMemFileSystem(FileLikeSystem):
    def __init__(self):
        self._files = {}

    @property
    def template(self) -> str:
        """Template to generate the full filename"""
        return "{}"

    @template.setter
    def template(self, value: str):
        """Change template for locating the files"""
        ...

    def open(self, filename: str, mode: str = "r") -> InMemFile:
        content = None
        if filename in self._files:
            content = self._files[filename]._content

        self._files[filename] = InMemFile(filename, mode, content)
        return self._files[filename]

    async def rm(self, pattern: str):
        fns = await self.ls(pattern)
        for fn in fns:
            self._files.pop(fn)

    async def ls(self, pattern: str = "*"):
        return [
            value
            for value in self._files.keys()
            if pattern == "*" or fnmatch(value, pattern)
        ]
